from __future__ import annotations

from typing import Dict
from typing import List

TYPES: List[Dict] = [
    {
        'type': 'BREAKING CHANGE',
        'description': 'Changes that are not backward-compatibles',
        'heading': 'Breaking changes',
        'emoji': '🚨',
        'bump': 'MAJOR',
        'regex': r'BREAKING[\-\ ]CHANGE',
        'question': False,  # Breaking changes have a dedicated question
        'changelog': True,
    },
    {
        'type': 'feat',
        'description': 'A new feature',
        'heading': 'New features',
        'emoji': '✨',
        'bump': 'MINOR',
        'changelog': True,
    },
    {
        'type': 'fix',
        'description': 'A bug fix',
        'heading': 'Bug fixes',
        'emoji': '🐛',
        'bump': 'PATCH',
        'changelog': True,
    },
    {
        'type': 'change',
        'description': 'A change made to the codebase.',
        'heading': 'Changes',
        'emoji': '🏗️',
        'bump': 'PATCH',
        'changelog': True,
    },
    {
        'type': 'docs',
        'description': 'Documentation only change',
        'heading': 'Documentation',
        'emoji': '📖',
        'bump': 'PATCH',
        'changelog': True,
    },
    {
        'type': 'test',
        'description': 'Adding missing or correcting existing tests',
        'heading': 'Testing',
        'emoji': '🚦',
        'bump': 'PATCH',
        'changelog': False,
    },
    {
        'type': 'ci',
        'description': 'Changes to CI configuration files and scripts',
        'heading': 'CI and project settings',
        'emoji': '⚙️',
        'bump': 'PATCH',
        'changelog': False,
    },
    {
        'type': 'refactor',
        'description': 'A changeset neither fixing a bug nor adding a feature',
        'heading': 'Refactoring',
        'emoji': '🔧',
        'bump': 'PATCH',
        'changelog': False,
    },
    {
        'type': 'revert',
        'description': 'Revert one or more commits',
        'heading': 'Reverted',
        'emoji': '🔙',
        'bump': 'PATCH',
        'changelog': False,
    },
    {
        'type': 'remove',
        'description': 'Removing code or files',
        'heading': 'Removals',
        'emoji': '🗑️',
        'bump': 'PATCH',
        'changelog': True,
    },
]

BUMP_MESSAGE: str = 'change(bump): release $current_version → $new_version [skip-ci]'



from __future__ import annotations

import re
import sys

from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from functools import cached_property
from functools import total_ordering
from typing import Dict
from typing import List
from typing import Literal
from typing import Union

import defaults

from commitizen.config import read_cfg
from commitizen.defaults import Settings

RE_HTTP = re.compile(r'(?P<server>https?://.+)/(?P<repository>[^/]+/[^/]+/?)')
INCREMENT = Literal['MAJOR', 'MINOR', 'PATCH']


@dataclass
@total_ordering
class CommitType:
    type: str  # Key used as type in the commit header.
    description: str  # A human-readable description of the type.
    heading: Union[str, None] = None  # The resulting heading in the changelog for this type.
    emoji: Union[str, None] = None  # An optional emoji representing the type.
    changelog: bool = True  # Whether this type should appear in the changelog or not.
    question: bool = True  # Whether this type should appear in the question choices.
    bump: Union[INCREMENT, None] = None  # Specifies the version bump category (e.g., 'MAJOR', 'MINOR', 'PATCH').
    regex: Union[str, None] = None  # An optional regular expression matching this type.

    def __str__(self) -> str:
        return self.type

    def __hash__(self):
        return hash(self.type)

    def __eq__(self, other) -> bool:
        return self._compare_key(other) == 0

    def __lt__(self, other) -> bool:
        return self._compare_key(other) < 0

    def _compare_key(self, other) -> int:
        """Helper method for comparing with either CommitType or str."""
        if isinstance(other, CommitType):
            other_type = other.type.lower()
        elif isinstance(other, str):
            other_type = other.lower()
        else:
            return NotImplemented
        return (self.type.lower() > other_type) - (self.type.lower() < other_type)

    @classmethod
    def from_dict(cls, data: Dict) -> CommitType:
        """Creates a CommitType instance from a dictionary."""
        valid_fields = {f.name for f in fields(cls)}
        filtered_data = {k: v for k, v in data.items() if k in valid_fields}
        return cls(**filtered_data)

    @classmethod
    def from_list(cls, lst: List[Dict]) -> List[CommitType]:
        """Creates a list of CommitType instances from a list of dictionaries."""
        return [cls.from_dict(d) for d in lst]


class CzEspressifSettings(Settings):
    types: Union[list[dict], None]  # The list of accepted types
    extra_types: Union[list[dict], None]  # A list of additional types (permit addition without losing defaults)
    github: Union[str, None]
    gitlab: Union[str, None]
    jira_url: Union[str, None]
    jira_prefixes: Union[list[str], None]
    order_by_scope: Union[bool, None]
    group_by_scope: Union[bool, None]
    include_unreleased: Union[bool, None]


@dataclass
class CzEspressifConfig:
    settings: CzEspressifSettings = field(default_factory=lambda: read_cfg().settings)

    @property
    def types(self) -> list[CommitType]:
        return CommitType.from_list(self.settings.get('types', defaults.TYPES))

    @property
    def extra_types(self) -> list[CommitType]:
        return CommitType.from_list(self.settings.get('extra_types', []))

    @cached_property
    def known_types(self) -> list[CommitType]:
        return self.types + self.extra_types

    @cached_property
    def github(self) -> Union[str, None]:
        repository = self.settings.get('github')
        if not repository:
            return None
        match = RE_HTTP.match(repository)
        return match.group('repository') if match else repository

    @cached_property
    def github_url(self) -> str:
        repository = self.settings.get('github')
        if repository:
            match = RE_HTTP.match(repository)
            if match:
                return match.group('server')
        return 'https://github.com'

    @cached_property
    def gitlab(self) -> Union[str, None]:
        repository = self.settings.get('gitlab')
        if not repository:
            return None
        match = RE_HTTP.match(repository)
        return match.group('repository') if match else repository

    @cached_property
    def gitlab_url(self) -> str:
        repository = self.settings.get('gitlab')
        if repository:
            match = RE_HTTP.match(repository)
            if match:
                return match.group('server')
        return 'https://gitlab.com'

    @cached_property
    def jira_url(self) -> Union[str, None]:
        return self.settings.get('jira_url')

    @cached_property
    def jira_prefixes(self) -> list[str]:
        return self.settings.get('jira_prefixes', [])

    @property
    def incremental(self) -> bool:
        return '--incremental' in sys.argv

    @property
    def order_by_scope(self) -> bool:
        return self.settings.get('order_by_scope') or False

    @property
    def include_unreleased(self) -> bool:
        return self.settings.get('include_unreleased', True)

    @property
    def use_emoji(self) -> bool:
        return self.settings.get('use_emoji', True)

    # @property
    # def bump_message(self) -> str:
    #     if not



import itertools

from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Union

from commitizen.cz import exceptions
from commitizen.cz.base import BaseCommitizen
from commitizen.cz.base import BaseConfig
from commitizen.defaults import Questions
from config import INCREMENT
from config import CommitType
from config import CzEspressifConfig
from example import build_example
from jinja2 import PackageLoader


def required_validator(answer: str, msg: str = '') -> str:
    if not answer:
        raise exceptions.AnswerRequiredError(msg)
    return answer


def multiple_line_breaker(answer: str, sep: str = '|') -> str:
    return '\n'.join(line.strip() for line in answer.split(sep) if line)


def parse_scope(text: str) -> str:
    if not text:
        return ''
    scope = text.strip().split()
    if len(scope) == 1:
        return scope[0]
    return '-'.join(scope)


def parse_subject(text: str) -> str:
    if isinstance(text, str):
        text = text.strip('.').strip()
    return required_validator(text, msg='Subject is required.')


class CzPluginEspressif(BaseCommitizen):  # pylint: disable=abstract-method
    template = 'CHANGELOG.md.j2'
    template_loader = PackageLoader('czespressif', 'templates')

    def __init__(self, config: BaseConfig):
        super().__init__(config)
        self.cze_config = CzEspressifConfig(config.settings)

    @property
    def known_types(self) -> Dict[str, CommitType]:
        return dict(
            itertools.chain(
                ((t.type, t) for t in self.cze_config.known_types),
            )
        )

    @property
    def re_known_types(self) -> str:
        return '|'.join(t.regex or k for k, t in self.known_types.items())

    @property
    def changelog_pattern(self) -> str:
        re_known_types = '|'.join(k for k, t in self.known_types.items())
        return rf'\A({re_known_types})(\(.+\))?(!)?'

    @property
    def change_type_map(self) -> Dict[str, str]:
        return dict(
            itertools.chain(
                ((t.type, f'{t.emoji + " " if self.cze_config.use_emoji and t.emoji else ""}{t.heading}') for t in self.cze_config.known_types if t.heading),
            )
        )

    @property
    def change_type_order(self) -> List[str]:
        return [
            f'{commit_type.emoji} {commit_type.heading}' if self.cze_config.use_emoji and commit_type.emoji else commit_type.heading
            for commit_type in self.cze_config.known_types
            if commit_type.changelog and commit_type.heading
        ]

    @property
    def bump_pattern(self) -> str:
        """Regex to extract information from commit (subject and body)"""
        re_types = '|'.join(t.regex or k for k, t in self.known_types.items() if t.bump)
        return rf'^((({re_types})(\(.+\))?(!)?)|\w+!):'

    @property
    def bump_map(self) -> Dict[str, INCREMENT]:
        """
        Mapping the extracted information to a SemVer increment type (MAJOR, MINOR, PATCH)
        """
        return OrderedDict(
            (
                (r'^.+!$', 'MAJOR'),
                *((rf'^{t.regex or k}', t.bump) for k, t in self.known_types.items() if t.bump),
            )
        )

    @property
    def bump_map_major_version_zero(self) -> Dict[str, INCREMENT]:
        return OrderedDict((pattern, increment.replace('MAJOR', 'MINOR')) for pattern, increment in self.bump_map.items())

    @property
    def commit_parser(self) -> str:
        return (
            rf'(?P<change_type>{self.re_known_types})'
            r'(?:\((?P<scope>[^()\r\n]*)\))?'  # Optional scope
            r'(?P<breaking>!)?:\s'
            r'(?P<message>.*)'  # Required message
            r'(?:\n{2,}(?P<body>.*))?'  # Optional body
            r'(?:\n{2,}(?P<footer>.*))?'  # Optional footer
        )

    def questions(self) -> Questions:
        """Questions regarding the commit message."""
        questions: List[Dict[str, Any]] = [
            {
                'type': 'list',
                'name': 'type',
                'message': 'Select the type of change you are committing',
                'choices': [
                    {
                        'value': t.type,
                        'name': f'{t.emoji + " " if self.cze_config.use_emoji and t.emoji else ""}{t.type}: {t.description}',
                    }
                    for t in self.cze_config.known_types
                    if t.question
                ],
            },
            {
                'type': 'input',
                'name': 'scope',
                'message': 'Scope. Define Could be anything specifying place of the commit change (wifi, bt, core):\n',
                'filter': parse_scope,
            },
            {
                'type': 'input',
                'name': 'subject',
                'filter': parse_subject,
                'message': 'Subject. Concise description of the changes. Imperative, lower case and no period:\n',
            },
            {
                'type': 'input',
                'name': 'body',
                'message': 'Provide additional contextual information - commit message BODY: (press [enter] to skip)\n',
                'filter': multiple_line_breaker,
            },
            {
                'type': 'confirm',
                'name': 'is_breaking_change',
                'message': 'Is this a BREAKING CHANGE?\n',
                'default': False,
            },
            {
                'when': lambda x: x['is_breaking_change'],
                'type': 'input',
                'name': 'breaking_change',
                'message': 'Breaking changes. Details the breakage:\n',
                'filter': multiple_line_breaker,
            },
            {
                'type': 'input',
                'name': 'footer',
                'message': 'Footer. Information about Breaking Changes and reference issues that this commit impacts: (press [enter] to skip)\n',
                'filter': multiple_line_breaker,
            },
        ]
        return questions

    def message(self, answers: Dict[str, Any]) -> str:
        prefix = answers['type']
        scope = answers['scope']
        subject = answers['subject']
        body = answers['body']
        is_breaking_change = answers['is_breaking_change']
        breaking_change = answers.get('breaking_change')
        footer = answers.get('footer', '')
        extra = ''

        if scope:
            scope = f'({scope})'

        if is_breaking_change and not breaking_change:
            scope += '!'

        if body:
            body = f'\n\n{body}'

        if is_breaking_change and breaking_change:
            footer = f'BREAKING CHANGE: {breaking_change}\n{footer}'

        if footer:
            footer = f'\n\n{footer}'

        message = f'{prefix}{scope}: {subject}{extra}{body}{footer}'

        return message

    def changelog_message_builder_hook(self, parsed_message: Dict[str, Any], commit: Any) -> Union[Dict[str, Any], Iterable[Dict[str, Any]], None]:  # pylint: disable=unused-argument
        # Remap breaking changes type
        if parsed_message.get('breaking'):
            parsed_message['change_type'] = 'BREAKING CHANGE'

        # Get the full list of types to include from pyproject.toml
        included_in_changelog = self.cze_config.settings.get('types_in_changelog', [])

        # If the list is empty (i.e., not set), include all default types
        if not included_in_changelog:
            commit_type = self.known_types.get(parsed_message['change_type'])
            if commit_type and commit_type.changelog:
                return parsed_message
            return None

        # Filter out all commit types not explicitly included in `types_in_changelog`
        if parsed_message['change_type'] not in included_in_changelog:
            return None
        return parsed_message

    def changelog_hook(self, full: str, partial: Union[str, None]) -> str:
        """Process resulting changelog to keep 1 empty line at the end of the file."""
        changelog = partial or full
        return changelog.rstrip() + '\n'

    def example(self) -> str:
        examples = build_example(types=self.cze_config.types, extra_types=self.cze_config.extra_types)
        return examples

    def schema(self) -> str:
        return '<type>(<scope>): <subject>\n' '<BLANK LINE>\n' '<body>\n' '<BLANK LINE>\n' '(BREAKING CHANGE: <breaking changes>)*\n' '(<footers>)*'

    def schema_pattern(self) -> str:
        pattern = (
            r'(?s)'  # To explicitly make . match new line
            rf'({self.re_known_types})'  # type
            r'(\(\S+\))?!?:'  # scope
            r'( [^\n\r]+)'  # subject
            r'((\n\n.*)|(\s*))?$'
        )
        return pattern

    @property
    def template_extras(self) -> Dict[str, Any]:
        return {'config': self.cze_config, 'settings': self.cze_config.settings}

    def info(self) -> str:
        return (
            'This is a custom Commitizen plugin for Espressif Systems. \n'
            'It is mainly used to generate consistent and beautiful changelog with minimum settings from target project. \n'
            'Can be used for creating commit messages, that passes Espressif standards. \n'
            'Customization from target project is possible, check this plugin GitHub repo https://github.com/espressif/cz-plugin-espressif\n\n'
            'More about internal standards, commit rules, and guidelines can be found in the https://github.com/espressif/standards.'
        )


discover_this = CzPluginEspressif  # pylint: disable=invalid-name



from typing import List

from config import CommitType

# ANSI escape codes for colors and bold text
YELLOW = '\033[93m'
GREEN = '\033[92m'
RED = '\033[91m'
BOLD = '\033[1m'
LIGHT_BLUE = '\033[96m'
PURPLE = '\033[35m'
BLUE = '\033[94m'
RESET = '\033[0m'


def build_example(types: List[CommitType], extra_types: List[CommitType]) -> str:
    # Exclude 'BREAKING CHANGE' from the known types list
    filtered_types = [t for t in types if t.type != 'BREAKING CHANGE']

    # Format the known types section with two spaces of indentation
    types_formatted = '\n'.join(f'  {GREEN}{t.type}{RESET}: {t.description}' for t in filtered_types)
    extra_types_formatted = '\n'.join(f'  {BLUE}{t.type}{RESET}: {t.description}' for t in extra_types)
    if extra_types_formatted:
        extra_types_formatted += f'  {BLUE}<--- non standard type added by extra_types{RESET}'

    # Example guide with colors and bold text
    example_guide = f"""
---

{BOLD}{YELLOW}Commit message schema:{RESET}

    {YELLOW}<type>(<scope>): <subject>{RESET}
        <... empty line ...>
    {YELLOW}<body>{RESET}
        <... empty line ...>
    {YELLOW}(BREAKING CHANGE: <breaking changes>)*{RESET}
    {YELLOW}(<footers>)*{RESET}

---

{BOLD}{GREEN}Commit types in this project:{RESET}

{types_formatted}
{extra_types_formatted}

---

{BOLD}{LIGHT_BLUE}Short commit messages (one line):{RESET}

    {LIGHT_BLUE}fix{RESET}: fix something (present or imperative, no period)
    {LIGHT_BLUE}fix(package){RESET}: fix something
    {LIGHT_BLUE}feat(name){RESET}: add a new feature
    {LIGHT_BLUE}refactor(package){RED}!{RESET}: this refactoring {RED}breaks the thing{RESET}

---

{BOLD}{BLUE}Full commit messages with body and footer:{RESET}

    {BLUE}feat(something){RESET}: add something new (present/imperative, no period)

    Here optionally comes the details - this is commit message body.
    It can be multiline, hyphens, asterisks, are okay.

    {RED}BREAKING CHANGE: It breaks something!{RESET}
    {BLUE}Closes https://github.com/espressif/<project>/issues/<issue_number>{RESET}

---

{BOLD}{PURPLE}More info:{RESET}

- https://www.conventionalcommits.org/en/v1.0.0/#specification
- https://github.com/{PURPLE}espressif/conventional-precommit-linter{RESET}
- https://github.com/{PURPLE}espressif/standards{RESET}
    """

    return example_guide
